//=============================================================================
// Ring Buffer Class Library
// Copyright(c)1999 Yamamoto Takanori.
//=============================================================================
#include <stdio.h>  // for fprintf()
#include <stdlib.h> // for malloc(), free()
#include "ezring.h"

//CONSTRUCTOR
ezring::ezring(int size)
{
  ring_ptr_head = ring_ptr_tail = 0;
  ring_bufsize = size;
  ring_buff = (char *)malloc(ring_bufsize);
}

//DESTRUCTOR
ezring::~ezring()
{
  free(ring_buff);
}

int ezring::flush(void)
{
  ring_ptr_head = ring_ptr_tail = 0;
  return 0;
}

int ezring::append(const char *str, int len)
{
  int n;
  for(n=0; n<len; n++){
    ring_buff[(ring_ptr_tail + n) % ring_bufsize] = str[n];
  }
  ring_ptr_tail = (ring_ptr_tail + len) % ring_bufsize;
  return 0;
}

int ezring::extract(char *str)
{
  fprintf(stderr, "extract() is obsoleted. use more safety functinon nextract().\n");
  int n = 0;
  while(ring_ptr_head != ring_ptr_tail){
    str[n++] = ring_buff[ring_ptr_head];
    ring_ptr_head = (ring_ptr_head + 1) % ring_bufsize;
  }
  return n;
}

int ezring::nextract(char *str, int len)
{
  int n = 0;
  while(len-- && (ring_ptr_head != ring_ptr_tail)){
    str[n++] = ring_buff[ring_ptr_head];
    ring_ptr_head = (ring_ptr_head + 1) % ring_bufsize;
  }
  return n;
}

int ezring::peep(char *str)
{
  fprintf(stderr, "peep() is obsoleted. use more safety functinon npeep().\n");
  int ptr = ring_ptr_head;
  int n = 0;
  while(ptr != ring_ptr_tail){
    str[n++] = ring_buff[ptr];
    ptr = (ptr + 1) % ring_bufsize;
  }
  return n;
}

int ezring::npeep(char *str, int len)
{
  int ptr = ring_ptr_head;
  int n = 0;
  while(len-- && ptr != ring_ptr_tail){
    str[n++] = ring_buff[ptr];
    ptr = (ptr + 1) % ring_bufsize;
  }
  return n;
}

int ezring::length(void)
{
  int len;
  len = (ring_ptr_tail - ring_ptr_head);
  if(len < 0){len += ring_bufsize;}
  return len;
}

int ezring::strncmp(const char *str, int len)
{
  int n;
  if(this->length() < len){return -1;}
  for(n=0; n<len; n++){
    if(str[n] != ring_buff[(ring_ptr_head + n) % ring_bufsize]){return -1;}
  }
  return 0;
}

int ezring::getbufsize(void)
{
  return(ring_bufsize);
}

int ezring::setbufsize(int size)
{
  free(ring_buff);
  ring_ptr_head = ring_ptr_tail = 0;
  ring_bufsize = size;
  ring_buff = (char *)malloc(ring_bufsize);
  return 0;
}


