#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>

#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <setjmp.h>

#include "topwindow.h"
#include "pixmap.h"
#include "mystyle.h"
#include "../tipo.h"
#include "../tipoware.h"

//G.Variables
jmp_buf env;

const char *sDev = DEVICE;
const int nBps = B19200;

static void destroy(){gtk_exit(0);}

void local_mkdir(GtkWidget *widget, gpointer data);
void local_delete(GtkWidget *widget, gpointer data);
void local_refresh(GtkWidget *widget, gpointer data);
void local_changedir(GtkWidget *widget, gint row, gint column, GdkEventButton *e, gpointer data);

//void remote_delete(GtkWidget *widget, gpointer data);
void remote_refresh(GtkWidget *widget, gpointer data);

void gtipowaretimeout(void){
  fprintf(stderr, "serial timeout occurred.\n");
  longjmp(env, 0);
}


//////////////////
//CALLBACK_LOCAL//
//////////////////

void local_mkdir(GtkWidget *widget, gpointer data){
  local_refresh(widget, data);
}

void local_delete(GtkWidget *widget, gpointer data){
  GList *list;
  int num,j;
  list = GTK_CLIST(data)->selection;
  if(list==NULL){return;}//NODATA
  if((num=g_list_length(list))<1){return;}//NODATA
  for(j=0;list!=NULL;list=list->next){
    int n=(int)list->data;
    //begin_process
    gchar *ptr, *ptr2;
    gtk_clist_get_text(GTK_CLIST(data), n, 2, &ptr2);
    gtk_clist_get_text(GTK_CLIST(data), n, 1, &ptr);
    if(strcmp(ptr2, "<DIR>")){
      unlink(ptr);
    }else{
      if(strcmp(ptr, "..")){
        rmdir(ptr);
      }
    }
  }
  local_refresh(widget, data);
}

void local_refresh(GtkWidget *widget, gpointer data){
  DIR *dp;
  struct dirent *entry;
  struct stat statbuf;

  if((dp=opendir("./"))==NULL){
    fprintf(stderr, "cannot open directory");
    return;
  }

  char dir[1024]; getcwd(dir, sizeof(dir));
  //  puts(dir);

  gtk_clist_clear(GTK_CLIST(data));

  while((entry=readdir(dp))!=NULL){
    stat(entry->d_name, &statbuf);
    if(S_ISDIR(statbuf.st_mode)){
      if(strcmp(entry->d_name, ".")==0){continue;}
      gchar *items[4] = {"", entry->d_name, "<DIR>", ""};
      int n=gtk_clist_append(GTK_CLIST(data), items);
      gtk_clist_set_pixmap(GTK_CLIST(data), n, 0, pixmap_folder, pixmap_folder_mask);
    }else{
      char *sSize; sSize=(char *)malloc(statbuf.st_size/10+1); sprintf(sSize, "%ld", (long int)statbuf.st_size);
      char sDate[25];
      strncpy(sDate, (char *)ctime(&(statbuf.st_mtime)), 24); sDate[24]='\0';
      gchar *items[4] = {"", entry->d_name, sSize, sDate};
      int n=gtk_clist_append(GTK_CLIST(data), items);
      gtk_clist_set_pixmap(GTK_CLIST(data), n, 0, pixmap_doc, pixmap_doc_mask);
      free(sSize);
    }
  }
}

void local_changedir(GtkWidget *widget, gint row, gint column, GdkEventButton *e, gpointer data){
  if(e && (e->type==GDK_2BUTTON_PRESS || e->type==GDK_3BUTTON_PRESS)){
    GList *list;
    int num;
    list = GTK_CLIST(data)->selection;
    if(list==NULL){return;}//NODATA
    if((num=g_list_length(list))<1){return;}//NODATA
    int n=(int)list->data;
    //begin_process
    gchar *ptr;
    gtk_clist_get_text(GTK_CLIST(data), n, 2, &ptr);
    if(strcmp(ptr, "<DIR>")){return;}
    gtk_clist_get_text(GTK_CLIST(data), n, 1, &ptr);
    chdir(ptr);
    local_refresh(widget, data);
  }
}

///////////////////
//CALLBACK_REMOTE//
///////////////////

void remote_refresh(GtkWidget *widget, gpointer data){
  //TWLIST
    static int nCnt; nCnt=0;
    gtk_clist_clear(GTK_CLIST(data));
    TiPO tipo;
    if(tipo.open(sDev, nBps)){return; /* EXIT_DEVICEOPENERR */}
    tipo.setcallbackfunc_timeout(gtipowaretimeout);
    setjmp(env);
    if(nCnt++>=3){tipo.close(); return; /* EXIT_TIMEOUT */}
    ezlist heno;
    tipo.twlist(&heno);
    ezatom *ptr;
    heno.gofirst(&ptr);
    while(!heno.forward(&ptr)){
      char *str=heno.getdata(&ptr);
      char *s=NULL, *t=str;
      gchar *items[4] = {"", "", "", ""};//DUMMY
      int n=gtk_clist_append(GTK_CLIST(data), items);
      gtk_clist_set_pixmap(GTK_CLIST(data), n, 0, pixmap_doc, pixmap_doc_mask);
      for(char *u=t;*u!='\0';u++){if(*u=='\t'){*u='\0'; s=u+1; break;}}
      gtk_clist_set_text(GTK_CLIST(data), n, 3, t);//sDate
      t=s;
      for(char *u=t;*u!='\0';u++){if(*u=='\t'){*u='\0'; s=u+1; break;}}
      gtk_clist_set_text(GTK_CLIST(data), n, 2, t);//sSize
      t=s;
      for(char *u=t;*u!='\0';u++){if(*u=='\n'){*u='\0'; break;}}
      gtk_clist_set_text(GTK_CLIST(data), n, 1, t);//sFile
    }
    tipo.close();
}

void remote_delete(GtkWidget *widget, gpointer data){
  static GList *list;
  int num,j;
  list = GTK_CLIST(data)->selection;
  if(list==NULL){return;}//NODATA
  if((num=g_list_length(list))<1){return;}//NODATA
  for(j=0;list!=NULL;list=list->next){
    int n=(int)list->data;
    //begin_process
    gchar *ptr, *ptr2;
    gtk_clist_get_text(GTK_CLIST(data), n, 2, &ptr2);
    gtk_clist_get_text(GTK_CLIST(data), n, 1, &ptr);
    //TWDELE
    static int nCnt; nCnt=0;
    TiPO tipo;
    if(tipo.open(sDev, nBps)){return; /* EXIT_DEVICEOPENERR */}
    tipo.setcallbackfunc_timeout(gtipowaretimeout);
    setjmp(env);
    if(nCnt++>=3){tipo.close(); return; /* EXIT_TIMEOUT */}
    if(tipo.twdele(ptr)){tipo.close(); return; /* Anything wrong */}
    tipo.close();
  }
  remote_refresh(widget, data);
}

/////////////////////////
//CALLBACK_TRANSBUTTONS//
/////////////////////////

void transbuttons_twsend(GtkWidget *widget, gpointer data){
  static GList *list;
  int num,j;
  topwindow *top=(topwindow *)data;
  list = GTK_CLIST(top->local.listbox.clist_localfilelist)->selection;
  if(list==NULL){return;}//NODATA
  if((num=g_list_length(list))<1){return;}//NODATA
  for(j=0;list!=NULL;list=list->next){
    int n=(int)list->data;
    //begin_process
    gchar *ptr, *ptr2;
    gtk_clist_get_text(GTK_CLIST(top->local.listbox.clist_localfilelist), n, 2, &ptr2);
    gtk_clist_get_text(GTK_CLIST(top->local.listbox.clist_localfilelist), n, 1, &ptr);
    if(strcmp(ptr2, "<DIR>")){
      //TWSEND
      static int nCnt; nCnt=0;
      TiPO tipo;
      if(tipo.open(sDev, nBps)){return; /* EXIT_DEVICEOPENERR */}
      tipo.setcallbackfunc_timeout(gtipowaretimeout);
      setjmp(env);
      if(nCnt++>=3){tipo.close(); return; /* EXIT_TIMEOUT */}
      if(tipo.twsend(ptr)){tipo.close(); return; /* Anything wrong */}
    }else{
    }
  }
  local_refresh(widget, top->local.listbox.clist_localfilelist);
  remote_refresh(widget, top->remote.listbox.clist_remotefilelist);
}

void transbuttons_twrecv(GtkWidget *widget, gpointer data){
  static GList *list;
  int num,j;
  topwindow *top=(topwindow *)data;
  list = GTK_CLIST(top->remote.listbox.clist_remotefilelist)->selection;
  if(list==NULL){return;}//NODATA
  if((num=g_list_length(list))<1){return;}//NODATA
  for(j=0;list!=NULL;list=list->next){
    int n=(int)list->data;
    //begin_process
    gchar *ptr, *ptr2;
    gtk_clist_get_text(GTK_CLIST(top->remote.listbox.clist_remotefilelist), n, 2, &ptr2);
    gtk_clist_get_text(GTK_CLIST(top->remote.listbox.clist_remotefilelist), n, 1, &ptr);
    //TWRECV
    static int nCnt; nCnt=0;
    TiPO tipo;
    if(tipo.open(sDev, nBps)){return; /* EXIT_DEVICEOPENERR */}
    tipo.setcallbackfunc_timeout(gtipowaretimeout);
    setjmp(env);
    if(nCnt++>=3){tipo.close(); return; /* EXIT_TIMEOUT */}
    if(tipo.twrecv(ptr)){tipo.close(); return; /* Anything wrong */}
    tipo.close();
  }
  local_refresh(widget, top->local.listbox.clist_localfilelist);
  remote_refresh(widget, top->remote.listbox.clist_remotefilelist);
}

//////////////////////////
//CALLBACK_BOTTOMBUTTONS//
//////////////////////////

void bottombuttons_about(GtkWidget *widget, gpointer data){
  GtkWidget *dialog_about=NULL;
  GtkWidget *button;
  GtkWidget *label;
  GtkWidget *picture;


  //PIXMAP
  picture = gtk_pixmap_new(pixmap_portrait, pixmap_portrait_mask);
  gtk_widget_set_mystyle(picture);
  gtk_widget_show(picture);
 
  //LABEL
  label = gtk_label_new(" This is gTiPOWARE, version 0.12\nCopyright(c)1999 Yamamoto Takanori\n/ Tatyana <tatyana@miko.org>");
  gtk_widget_set_mystyle(label);
  gtk_widget_show(label);
  //BUTTON
  button = gtk_button_new_with_label(" OK ");
  gtk_widget_set_mystyle(button);
  gtk_widget_show(button);
  //DIALOG
  dialog_about = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(dialog_about), "About gTiPOWARE");
  gtk_container_border_width(GTK_CONTAINER(dialog_about),6);
  GtkWidget *hbox=gtk_hbox_new(FALSE,4);
  gtk_container_add(GTK_CONTAINER(dialog_about), hbox);
  gtk_widget_set_mystyle(hbox);
  gtk_widget_show(hbox);

  gtk_box_pack_start(GTK_BOX(hbox), picture, FALSE, FALSE, 10);
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 10);
  GtkWidget *vbox=gtk_vbox_new(FALSE,4);
  gtk_widget_set_mystyle(vbox);
  gtk_widget_show(vbox);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 5);
  gtk_box_pack_end(GTK_BOX(vbox), button, FALSE, FALSE, 5);

  //CALLBACK
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
  gtk_signal_connect(GTK_OBJECT(dialog_about), "delete_event", GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

  gtk_widget_set_mystyle(dialog_about);
  gtk_widget_show(dialog_about);

  gtk_grab_add(dialog_about);
  gtk_main();
  gtk_grab_remove(dialog_about);

  gtk_signal_handlers_destroy(GTK_OBJECT(dialog_about));
  gtk_widget_destroy(dialog_about);
}

void bottombuttons_notimp(GtkWidget *widget, gpointer data){
  GtkWidget *dialog_about=NULL;
  GtkWidget *button;
  GtkWidget *label;

  //LABEL
  label = gtk_label_new(" This function is not implemented yet.. ");
  gtk_widget_set_mystyle(label);
  gtk_widget_show(label);
  //BUTTON
  button = gtk_button_new_with_label(" OK ");
  gtk_widget_set_mystyle(button);
  gtk_widget_show(button);
  //DIALOG
  dialog_about = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(dialog_about), "gTiPOWARE Message Dialog");
  gtk_container_border_width(GTK_CONTAINER(dialog_about),6);
  GtkWidget *hbox=gtk_hbox_new(FALSE,4);
  gtk_container_add(GTK_CONTAINER(dialog_about), hbox);
  gtk_widget_set_mystyle(hbox);
  gtk_widget_show(hbox);

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 10);
  GtkWidget *vbox=gtk_vbox_new(FALSE,4);
  gtk_widget_set_mystyle(vbox);
  gtk_widget_show(vbox);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 5);
  gtk_box_pack_end(GTK_BOX(vbox), button, FALSE, FALSE, 5);

  //CALLBACK
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
  gtk_signal_connect(GTK_OBJECT(dialog_about), "delete_event", GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

  gtk_widget_set_mystyle(dialog_about);
  gtk_widget_show(dialog_about);

  gtk_grab_add(dialog_about);
  gtk_main();
  gtk_grab_remove(dialog_about);

  gtk_signal_handlers_destroy(GTK_OBJECT(dialog_about));
  gtk_widget_destroy(dialog_about);
}

void bottombuttons_exit(GtkWidget *widget, gpointer data){
  gtk_main_quit();
}

/////////////////
//MAIN FUNCTION//
/////////////////

int main(int argc, char *argv[]){
  topwindow top;

  //INIT
  gtk_set_locale();
  gtk_init(&argc, &argv);

  //CREATE_TOPLEVEL_WINDOW
  top.create();
  gtk_signal_connect(GTK_OBJECT(top.widget), "destroy", GTK_SIGNAL_FUNC(destroy), NULL);

  //CREATE_PIXMAP
  pixmap_create(top.widget);

  //CONNECT_CALLBACK_LOCALSYSTEM
  gtk_signal_connect(GTK_OBJECT(top.local.buttonbar.button_mkdir), "clicked", GTK_SIGNAL_FUNC(bottombuttons_notimp), top.local.listbox.clist_localfilelist);
  gtk_signal_connect(GTK_OBJECT(top.local.buttonbar.button_refresh), "clicked", GTK_SIGNAL_FUNC(local_refresh), top.local.listbox.clist_localfilelist);
  gtk_signal_connect(GTK_OBJECT(top.local.buttonbar.button_delete), "clicked", GTK_SIGNAL_FUNC(local_delete), top.local.listbox.clist_localfilelist);
  gtk_signal_connect(GTK_OBJECT(top.local.listbox.clist_localfilelist), "select_row", GTK_SIGNAL_FUNC(local_changedir), top.local.listbox.clist_localfilelist);

  //CONNECT_CALLBACK_REMOTETIPO
  gtk_signal_connect(GTK_OBJECT(top.remote.buttonbar.button_mkdir), "clicked", GTK_SIGNAL_FUNC(bottombuttons_notimp), top.remote.listbox.clist_remotefilelist);
  gtk_signal_connect(GTK_OBJECT(top.remote.buttonbar.button_refresh), "clicked", GTK_SIGNAL_FUNC(remote_refresh), top.remote.listbox.clist_remotefilelist);
  gtk_signal_connect(GTK_OBJECT(top.remote.buttonbar.button_delete), "clicked", GTK_SIGNAL_FUNC(remote_delete), top.remote.listbox.clist_remotefilelist);

  //CONNECT_CALLBACK_TRANSBUTTON
  gtk_signal_connect(GTK_OBJECT(top.trans.button_twsend), "clicked", GTK_SIGNAL_FUNC(transbuttons_twsend), &top);
  gtk_signal_connect(GTK_OBJECT(top.trans.button_twrecv), "clicked", GTK_SIGNAL_FUNC(transbuttons_twrecv), &top);

  //CONNECT_CALLBACK_BOTTOMBUTTON
  gtk_signal_connect(GTK_OBJECT(top.bottom.button_connect), "clicked", GTK_SIGNAL_FUNC(remote_refresh), top.remote.listbox.clist_remotefilelist);
  gtk_signal_connect(GTK_OBJECT(top.bottom.button_logwnd), "clicked", GTK_SIGNAL_FUNC(bottombuttons_notimp), NULL);
  gtk_signal_connect(GTK_OBJECT(top.bottom.button_help), "clicked", GTK_SIGNAL_FUNC(bottombuttons_notimp), NULL);
  gtk_signal_connect(GTK_OBJECT(top.bottom.button_options), "clicked", GTK_SIGNAL_FUNC(bottombuttons_notimp), NULL);
  gtk_signal_connect(GTK_OBJECT(top.bottom.button_about), "clicked", GTK_SIGNAL_FUNC(bottombuttons_about), NULL);
  gtk_signal_connect(GTK_OBJECT(top.bottom.button_exit), "clicked", GTK_SIGNAL_FUNC(bottombuttons_exit), NULL);


  //REFRESH
  local_refresh(top.local.buttonbar.button_refresh, top.local.listbox.clist_localfilelist);

  gtk_main();
  return 0;
}
 






