//=============================================================================
// Nihongo Code Convert Library
// Copyright(c)1999 Yamamoto Takanori.
//=============================================================================
#include "jcode.h"

#define BYTE unsigned char
#define WORD unsigned short

#define WORD_U(d) (((d)>>8)&0xff)
#define WORD_L(d) ((d)&0xff)

// Ref.
// http://joe.ash.or.jp/tips/code/index.htm, etc.

int issjis1(BYTE c){
  return (0x81<=c && c<=0x9f || 0xe0<=c && c<=0xfc);
}
int issjis2(BYTE c){
  return (0x40<=c && c<=0x7e || 0x80<=c && c<=0xfc);
}

int issjis_jiskana(BYTE c){
  return (0xa1<=c && c<=0xdf);
}

#if 0
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","","",
    "","","","","","","",""
#endif

WORD sjis_jiskana2jiskanji(BYTE c){
  static const WORD han2zen[64]={
  0x8140, 0x8142, 0x8175, 0x8176, 0x8141, 0x8145, 0x8392, 0x8340,
  0x8342, 0x8344, 0x8346, 0x8348, 0x8383, 0x8385, 0x8387, 0x8362,
  0x815b, 0x8341, 0x8343, 0x8345, 0x8347, 0x8349, 0x834a, 0x834c,
  0x834e, 0x8350, 0x8352, 0x8354, 0x8356, 0x8358, 0x835a, 0x835c,
  0x835e, 0x8360, 0x8363, 0x8365, 0x8367, 0x8369, 0x836a, 0x836b,
  0x836c, 0x836d, 0x836e, 0x8371, 0x8374, 0x8377, 0x837a, 0x837d,
  0x837e, 0x8380, 0x8381, 0x8382, 0x8384, 0x8386, 0x8388, 0x8389,
  0x838a, 0x838b, 0x838c, 0x838d, 0x838f, 0x8393, 0x814a, 0x814b};
  return han2zen[(c-0xa0)];
}

WORD euc_jiskana2jiskanji(WORD c){
  static const WORD han2zen[64]={
  0xa1a1, 0xa1a3, 0xa1d6, 0xa1d7, 0xa1a2, 0xa1a6, 0xa5f2, 0xa5a1,
  0xa5a3, 0xa5a5, 0xa5a7, 0xa5a9, 0xa5e3, 0xa5e5, 0xa5e7, 0xa5c3,
  0xa1bc, 0xa5a2, 0xa5a4, 0xa5a6, 0xa5a8, 0xa5aa, 0xa5ab, 0xa5ad,
  0xa5af, 0xa5b1, 0xa5b3, 0xa5b5, 0xa5b7, 0xa5b9, 0xa5bb, 0xa5bd,
  0xa5bf, 0xa5c1, 0xa5c4, 0xa5c6, 0xa5c8, 0xa5ca, 0xa5cb, 0xa5cc,
  0xa5cd, 0xa5ce, 0xa5cf, 0xa5d2, 0xa5d5, 0xa5d8, 0xa5db, 0xa5de,
  0xa5df, 0xa5e0, 0xa5e1, 0xa5e2, 0xa5e4, 0xa5e6, 0xa5e8, 0xa5e9,
  0xa5ea, 0xa5eb, 0xa5ec, 0xa5ed, 0xa5ef, 0xa5f3, 0xa1ab, 0xa1ac};
  return han2zen[(c-0x8ea0)];
}
unsigned char *streuc_jiskana2jiskanji(const unsigned char *str, unsigned char *buff){
  int n=0;
  WORD w;
  while(str[n]!='\0'){
    if(iseuc_jiskana1(str[n])&&iseuc_jiskana2(str[n+1])){
      w=euc_jiskana2jiskanji(byte2word(str[n],str[n+1]));
      buff[n]=WORD_U(w); buff[n+1]=WORD_L(w);
      n+=2;
     }else{
      buff[n]=str[n];
      n++;
    }
  }
  buff[n]='\0';
  return buff;
}
unsigned char *strsjis_jiskana2jiskanji(const unsigned char *str, unsigned char *buff){
  int m=0, n=0;
  WORD w;
  while(str[n]!='\0'){
    if(issjis1(str[n])&&issjis2(str[n+1])){
      buff[m]=str[n]; buff[m+1]=str[n+1];
      m+=2; n+=2;
     }else if(issjis_jiskana(str[n])){
      w=sjis_jiskana2jiskanji(str[n]);
      buff[m]=WORD_U(w); buff[m+1]=WORD_L(w);
      m+=2; n++;
     }else{
      buff[m]=str[n];
      m++; n++;
    }
  }
  buff[m]='\0';
  return buff;
}

int isjis1(BYTE c){
  return (0x21<=c && c<=0x7e); //JIS KANJI
}
int isjis2(BYTE c){
  return (0x21<=c && c<=0x7e); //JIS KANJI
}
int iseuc1(BYTE c){
  return (0x8e<=c && c<=0xfe && c!=0x8f && c!=0xa0);
}
int iseuc2(BYTE c){
  return (0xa0<=c && c<=0xfe);
}
int iseuc_jiskana1(BYTE c){
  return (0x8e==c);
}
int iseuc_jiskana2(BYTE c){
  return (0xa1<=c && c<=0xdf);
}

WORD byte2word(BYTE u, BYTE l){
  return (u*0x100U + l);
}

WORD sjis2jis(WORD w){
  BYTE c1=WORD_U(w), c2=WORD_L(w);
  if(c1<0xa0){c1-=0x70;}
  else if(c1<0xf0){c1-=0xb0;}
  else{return byte2word(c1,c2);}
  if(c2>=0x80){c2--;}
  c1*=2;
  if(c2>=0x9e){c2-=0x5e;}
  else{c1--;}
  c2-=0x1f;
  return byte2word(c1,c2);
}

WORD jis2sjis(WORD w){
  BYTE c1=WORD_U(w), c2=WORD_L(w);
  if(c1&1){
    c2+=0x1f;
    if(c2>=0x7f){c2++;}
  }else{
    c2+=0x7e;
  }
  c1=(c1-0x21)/2+0x81;
  if(c1>=0xa0){c1+=0x40;}
  return(byte2word(c1,c2));
}

WORD jis2euc(WORD w){
  return w+0x8080; //JIS_KANJI
}

WORD euc2jis(WORD w){
  return w-0x8080; //JIS_KANJI
}

WORD sjis2euc(WORD w){
  return jis2euc(sjis2jis(w));
}

WORD euc2sjis(WORD w){
  return jis2sjis(euc2jis(w));
}

BYTE *strsjis2euc(const BYTE *str, BYTE *buff){
  int n=0;
  WORD w;
  while(str[n]!='\0'){
    if(issjis1(str[n])&&issjis2(str[n+1])){
      w=sjis2euc(byte2word(str[n],str[n+1]));
      buff[n]=WORD_U(w); buff[n+1]=WORD_L(w);
      n+=2;
    }else{
      buff[n]=str[n];
      n++;
    }
  }
  buff[n]='\0';
  return buff;
}

BYTE *streuc2sjis(const BYTE *str, BYTE *buff){
  int n=0;
  WORD w;
  while(str[n]!='\0'){
    if(iseuc1(str[n])&&iseuc2(str[n+1])){
      w=euc2sjis(byte2word(str[n],str[n+1]));
      buff[n]=WORD_U(w); buff[n+1]=WORD_L(w);
      n+=2;
    }else{
      buff[n]=str[n];
      n++;
    }
  }
  buff[n]='\0';
  return buff;
}

BYTE *strsjis2jis(const BYTE *str, BYTE *buff){
  int n=0;
  WORD w;
  while(str[n]!='\0'){
    if(issjis1(str[n])&&issjis2(str[n+1])){
      w=sjis2jis(byte2word(str[n],str[n+1]));
      buff[n]=WORD_U(w); buff[n+1]=WORD_L(w);
      n+=2;
    }else{
      buff[n]=str[n];
      n++;
    }
  }
  buff[n]='\0';
  return buff;
}

BYTE *strjis2sjis(const BYTE *str, BYTE *buff){
  int n=0;
  WORD w;
  while(str[n]!='\0'){
    if(isjis1(str[n])&&isjis2(str[n+1])){
      w=jis2sjis(byte2word(str[n],str[n+1]));
      buff[n]=WORD_U(w); buff[n+1]=WORD_L(w);
      n+=2;
    }else{
      buff[n]=str[n];
      n++;
    }
  }
  buff[n]='\0';
  return buff;
}

BYTE *streuc2jis(const BYTE *str, BYTE *buff){
  int n=0;
  WORD w;
  while(str[n]!='\0'){
    if(iseuc1(str[n])&&iseuc2(str[n+1])){
      w=euc2jis(byte2word(str[n],str[n+1]));
      buff[n]=WORD_U(w); buff[n+1]=WORD_L(w);
      n+=2;
    }else{
      buff[n]=str[n];
      n++;
    }
  }
  buff[n]='\0';
  return buff;
}

BYTE *strjis2euc(const BYTE *str, BYTE *buff){
  int n=0;
  WORD w;
  while(str[n]!='\0'){
    if(isjis1(str[n])&&isjis2(str[n+1])){
      w=jis2euc(byte2word(str[n],str[n+1]));
      buff[n]=WORD_U(w); buff[n+1]=WORD_L(w);
      n+=2;
    }else{
      buff[n]=str[n];
      n++;
    }
  }
  buff[n]='\0';
  return buff;
}

