//=============================================================================
// SII BrainPad(TM) TiPO Communication Class Library
// Copyright(c)1999 Yamamoto Takanori.
//=============================================================================
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <string.h>

#include "crc16.h"
#include "serial.h"
#include "jcode.h"
#include "tipo.h"
#include "exitstatus.h"

#define WORD_U(d) (((d)>>8)&0xff)
#define WORD_L(d) ((d)&0xff)

const char c_NUL = 0x00;
const char c_SOH = 0x01;
const char c_STX = 0x02;
const char c_ETX = 0x03;
const char c_EOT = 0x04;
const char c_ENQ = 0x05;
const char c_ACK = 0x06;
const char c_BEL = 0x07;
const char c_BS  = 0x08;
const char c_HT  = 0x09;
const char c_LF  = 0x0a;
const char c_VT  = 0x0b;
const char c_FF  = 0x0c;
const char c_CR  = 0x0d;
const char c_SO  = 0x0e;
const char c_SI  = 0x0f;

const char c_DLE = 0x10;
const char c_DC1 = 0x11;
const char c_DC2 = 0x12;
const char c_DC3 = 0x13;
const char c_DC4 = 0x14;
const char c_NAK = 0x15;
const char c_SYN = 0x16;
const char c_ETB = 0x17;
const char c_CAN = 0x18;
const char c_EM  = 0x19;
const char c_SUB = 0x1a;
const char c_ESC = 0x1b;

const char c_SPC = 0x20;

void timeoutisbad(void)
{
  fprintf(stderr, "serial timeout occurred.\n");
  exit(EXIT_TIMEOUT);
}

char *ptrfilename(char *s){
  char *ptr,*ptr2;
  ptr=ptr2=s;
  while(*ptr!='\0'){
    if(*ptr=='/'){ptr2=ptr+1;}
    ptr++;
  }
  return ptr2;
}

//CONSTRUCTOR
TiPO::TiPO(){
}

//DESTRUCTOR
TiPO::~TiPO(){
}

int TiPO::open(const char *dev, int bps)
{
  if(sio.open(dev, bps)){
    fprintf(stderr, "device open error - [%s]\n", dev);
    return(EXIT_DEVICEOPENERR);
  }
  sio.setbufsize(8192); // buffer size must be larger than 1K(YMODEM MAX BROCK SIZE)
  sio.settimeout(5000); // timeout is now 5 secs.
  sio.setcallbackfunc_timeout(timeoutisbad);
  return 0;
}

int TiPO::close()
{
  return sio.close();
}

long TiPO::twstat(void)
{
  long lFree;
  char *sTemp; sTemp = (char *)malloc(sio.getbufsize());

  sio.write(&c_CAN, 1); sio.write(&c_CAN, 1);
  sio.write("TWSTAT", 6);
  sio.read_until(sTemp, sio.getbufsize(), c_SOH);// ("TWGOOD" SOH)
  sio.read_until(sTemp, sio.getbufsize(), c_EOT);// ([FILESIZE] EOT)
  lFree = strtol(sTemp, NULL, 10);
  free(sTemp);

  return lFree;
}

int TiPO::twlist(ezlist *list)
{
  long lFSize;
  char sFTime[25];
  time_t tFTime;
  char *sTemp; sTemp = (char *)malloc(sio.getbufsize());
  char *sFileEUC; sFileEUC = (char *)malloc(sio.getbufsize());

  sio.write(&c_CAN, 1); sio.write(&c_CAN, 1);
  sio.write("TWLIST", 6);
  sio.read(sTemp, 6); // ("TWGOOD")
  while(1){
    sio.read(sTemp, 1);// SOH or EOT
    if(sTemp[0] == c_EOT){break;}

    //get sFileEUC
    sio.read_until(sTemp, sio.getbufsize(), c_NUL);// ([FILENAME] NUL)
    strsjis2euc((unsigned char *)sTemp, (unsigned char *)sFileEUC);

    //get lFSize
    sio.read_until(sTemp, sio.getbufsize(), c_SPC);// ([FILESIZE] SPC)
    lFSize = strtol(sTemp, NULL, 10);

    //get sFTime
    sio.read_until(sTemp, sio.getbufsize(), c_NUL);// ([TIMESTAMP] NUL)
    tFTime = mkunixtime(sTemp); strncpy(sFTime, ctime(&tFTime), 24);
    sFTime[24] = c_NUL;

    sprintf(sTemp, "%s\t%10ld\t%s", sFTime, lFSize, sFileEUC);
    list->append(sTemp, strlen(sTemp)+1);
  }
  free(sFileEUC); free(sTemp);
  return 0;
}

int TiPO::twsend(const char *sFileEUC, int flag_twveru)
{
  char *sTemp; sTemp = (char *)malloc(sio.getbufsize());
  size_t nFSize;
  unsigned short crc; char acCRC[2];
  char counter[2] = {0x00, 0xff};
  static FILE *fp; fp=NULL;

  //create_sFileSJIS
  int nFNLength = strlen(sFileEUC);
  char *sFileEUC_JISKANJI; sFileEUC_JISKANJI = (char *)malloc(nFNLength+1);
  streuc_jiskana2jiskanji((unsigned char *)sFileEUC,(unsigned char *)sFileEUC_JISKANJI);
  char *sFileSJIS; sFileSJIS = (char *)malloc(nFNLength+1);
  streuc2sjis((unsigned char *)sFileEUC_JISKANJI, (unsigned char *)sFileSJIS);
  strsjis_limitchrs(ptrfilename(sFileSJIS), 20);

  if((nFSize = fsize(sFileEUC))==0){
    fprintf(stderr, "file is empty - [%s]\n", sFileEUC);
    free(sTemp);
    free(sFileEUC_JISKANJI);
    free(sFileSJIS);
    return(EXIT_FILEOPENERR);
  }

  //begin_twsend
  sio.write(&c_CAN, 1); sio.write(&c_CAN, 1);
  if(flag_twveru){
    sio.write("TWVERU", 6);
  }else{
    sio.write("TWSEND", 6);
  }
  sio.read(sTemp, 6);// ("TWGOOD") or ("TWNOGD")
  if(istwgood(sTemp)){
    fprintf(stderr, "twsend error\n");
    free(sTemp);
    free(sFileEUC_JISKANJI);
    free(sFileSJIS);
    return (EXIT_TRANSMITERR);
  }
  sio.read(sTemp, 1);// ("C")

  //create_sBlock128
  do{
    sio.write(&c_SOH, 1);
    sio.write(counter, 2);
    //(sFileSJIS)
    char sBlock128[128]; for(int n=0;n<128;n++){sBlock128[n] = c_NUL;}
    int len1, len2;
    len1 = strncpy_until_wox(sBlock128, ptrfilename(sFileSJIS), 128, c_NUL); // NULL_TERMINATED
    //(sFSize)
    sprintf(sTemp, "%d", nFSize);
    len2 = strncpy_until_wox(sBlock128+(len1+1), sTemp, 128-(len1+1), c_NUL);
    sBlock128[(len1+1)+(len2)] = c_SPC; // SPC TERMINATED
    //(sFTime)
    time_t t; time(&t); mktipotime(t, sTemp); sTemp[10] = c_NUL;
    strncpy_until_wox(sBlock128+(len1+1)+(len2+1), sTemp, 128-(len1+1)-(len2+1), c_NUL); // NULL_TERMINATED


    //write_sBlock128
    sio.write(sBlock128, 128);
    crc = calc_crc16(sBlock128, 128);
    acCRC[0] = WORD_U(crc); acCRC[1] = WORD_L(crc);
    sio.write(acCRC, 2);

    sio.read(sTemp, 1); // (ACK or NAK or CAN)
    if(sTemp[0] == c_CAN){
      // FILENAME IS TOO LONG, TOO MANY CRC-ERROR, etc.
      fprintf(stderr, "twsend error - [%s]\n", sFileEUC_JISKANJI);
      free(sTemp);
      free(sFileSJIS);
      free(sFileEUC_JISKANJI);
      return(EXIT_TRANSMITERR2);
    }
  }while(sTemp[0] != c_ACK);

  free(sFileSJIS);
  free(sFileEUC_JISKANJI);

  sio.read(sTemp, 1); // ("C")

  if(fp!=NULL){fclose(fp);}
  if((fp=fopen(sFileEUC, "rb")) == NULL){
    fprintf(stderr, "file open error - [%s]\n", sFileEUC);
    free(sTemp);
    return(EXIT_FILEOPENERR);
  }

  //create_sBlock1024
  char sBlock1024[1024];
  while(nFSize){
    counter_inc(counter);
    if(nFSize >= 1024){
      fread(sBlock1024, 1024, 1, fp);
      nFSize-=1024;
    }else{
      for(int n=0;n<1024;n++){sBlock1024[n] = c_SUB;}
      fread(sBlock1024, nFSize, 1, fp);
      nFSize = 0;
    }
    do{
      sio.write(&c_STX,1);
      sio.write(counter, 2);
      sio.write(sBlock1024, 1024);
      crc = calc_crc16(sBlock1024, 1024);
      acCRC[0] = WORD_U(crc); acCRC[1] = WORD_L(crc);
      sio.write(acCRC, 2);
      sio.read(sTemp,1); // (ACK or NAK)
    }while(sTemp[0] != c_ACK); // CRC_ERR_CHK
  }while(nFSize);
  fclose(fp);
  sio.write(&c_EOT, 1);
  sio.read(sTemp, 1); // (ACK)

  free(sTemp);
  return 0;
}

int TiPO::twrecv(const char *sFileEUC)
{
  char *sTemp; sTemp = (char *)malloc(sio.getbufsize());
  char crc16[2];
  char counter[2];
  static FILE *fp; fp=NULL;
  int nFSize;

  //create_sFileSJIS
  int nFNLength = strlen(sFileEUC);
  char *sFileSJIS; sFileSJIS = (char *)malloc(nFNLength+1);
  streuc2sjis((unsigned char *)sFileEUC, (unsigned char *)sFileSJIS);

  //begin_twrecv
  sio.write(&c_CAN, 1); sio.write(&c_CAN, 1);
  sio.write("TWRECV", 6);
  sio.write(&c_SOH, 1); sio.write(sFileSJIS, strlen(sFileSJIS));
  sio.write(&c_EOT, 1);
  sio.read(sTemp, 6); // ("TWGOOD") or ("TWNOGD")
  if(istwgood(sTemp)){ // no such file?
    fprintf(stderr, "twrecv error - [%s]\n", sFileEUC);
    free(sTemp);
    free(sFileSJIS);
    return(EXIT_RECEIVEERR);
  }
  sio.write("C", 1);
  sio.read(sTemp, 1); // HEADER(SOH)

  free(sFileSJIS);

  //get_sBlock128
  while(1){
    char sBlock128[128];
    char sFileSJIS2[128], sFSize[128], sFTime[128];
    sio.read(counter, 2); // COUNTER(0x00 0xff)
    sio.read(sBlock128, 128); // (SOH [FILENAME] NUL [FILESIZE] SPC [TIMESTAMP] NUL ...)
    int len1, len2, len3;
    //(sFileSJIS2)
    len1 = strncpy_until_wox(sFileSJIS2, sBlock128+1, 128-1, c_NUL);
    //(sFSize)
    len2 = strncpy_until_wox(sFSize, sBlock128+1+(len1+1), 128-1-(len1+1), c_SPC);
    sFSize[len2] = c_NUL;
    nFSize = strtol(sFSize, NULL, 10);
    //(sFTime)
    len3 = strncpy_until_wox(sFTime, sBlock128+1+(len1+1)+(len2+1), 128-1-(len1+1)-(len2+1), c_NUL);
    //get_CRC
    sio.read(crc16, 2); // CRC16
    if(byte2word(crc16[0],crc16[1]) == calc_crc16(sBlock128, 128)){
      break;
    }
    fprintf(stderr, "CRC-Error detected in #%02X%02X\n",counter[0]&0xff,counter[1]&0xff);
    sio.write(&c_NAK, 1); // NAK
  }

  if(fp!=NULL){fclose(fp);}
  char *sFileEUC_NOSLASH; sFileEUC_NOSLASH = (char *)malloc(nFNLength+1);
  strslash2hyphen(sFileEUC, sFileEUC_NOSLASH);
  if((fp=fopen(sFileEUC_NOSLASH, "wb")) == NULL){
    fprintf(stderr, "file open error - [%s]\n", sFileEUC_NOSLASH);
    free(sTemp);
    free(sFileEUC_NOSLASH);
    return(EXIT_FILEOPENERR);
  }
  free(sFileEUC_NOSLASH);
  while(1){
    char sBlock1024[1024];
    sio.write(&c_ACK, 1); // ACK
    sio.read(sTemp, 1);// HEADER(STX or EOT)
    if(sTemp[0] == c_EOT){break;}
    while(1){
      sio.read(counter, 2); // COUNTER(0x01 0xfe, ...)
      sio.read(sBlock1024, 1024);
      sio.read(crc16, 2); // CRC16
      if(byte2word(crc16[0],crc16[1]) == calc_crc16(sBlock1024, 1024)){
        break;
      }
      fprintf(stderr, "CRC-Error detected in #%02X%02X\n",counter[0]&0xff,counter[1]&0xff);
      sio.write(&c_NAK, 1); // NAK
    }

    if(nFSize >= 1024){
      fwrite(sBlock1024, 1024, 1, fp);
      nFSize-=1024;
    }else{
      fwrite(sBlock1024, nFSize, 1, fp);
      nFSize=0;
    }
  }
  fclose(fp);
  free(sTemp);
  return 0;
}

int TiPO::twdele(const char *sFileEUC)
{
  char *sTemp; sTemp = (char *)malloc(sio.getbufsize());
  char *sFileSJIS;

  sFileSJIS = (char *)malloc(strlen(sFileEUC)+1);
  streuc2sjis((unsigned char *)sFileEUC, (unsigned char *)sFileSJIS);

  sio.write(&c_CAN, 1); sio.write(&c_CAN, 1);
  sio.write("TWDELE", 6);
  sio.write(&c_SOH, 1);
  sio.write(sFileSJIS, strlen(sFileSJIS));
  sio.write(&c_EOT, 1);

  free(sFileSJIS);

  sio.read(sTemp, 6);// ("TWGOOD") or ("TWNOGD")
  if(istwgood(sTemp)){
     fprintf(stderr, "twdele error - [%s]\n", sFileEUC);
     free(sTemp);
     return(EXIT_DELETEERR);
  }
  free(sTemp);
  return 0;
}

int TiPO::twvers(void)
{
  char *sTemp; sTemp = (char *)malloc(sio.getbufsize());
  int ret;

  sio.write(&c_CAN, 1); sio.write(&c_CAN, 1);
  sio.write("TWVERS", 6);
  sio.read(sTemp, 6); // ("TWGOOD")
  ret = istwgood(sTemp);
  free(sTemp);

  return ret; // 0 or -1
}

int TiPO::twveru(const char *sFileEUC)
{
  return twsend(sFileEUC, 1);
}

int TiPO::twexit(void)
{
  char *sTemp; sTemp = (char *)malloc(sio.getbufsize());
  sio.write(&c_CAN, 1); sio.write(&c_CAN, 1);
  sio.write("TWEXIT", 6);
  sio.read(sTemp, 6); // ("TWGOOD")
  free(sTemp);
  return 0;
}

int TiPO::setcallbackfunc_timeout(void (*f)(void)){
  sio.setcallbackfunc_timeout(f);
  return 0;
}


// misc functions.

char *mktipotime(time_t t, char *buf)
{
  int n=9;
  t+=32400; // Use Magic Number.
            // TiPO has a system-time adjusted to JST(GMT+9).

  while(n >= 0){
    buf[n] = '0'+(t%8);
    t /= 8;
    n--;
  }
  return buf;
}

time_t mkunixtime(const char *buf)
{
  int n;
  time_t t=0;
  for(n=0;n<10;n++){
    t*=8;
    t+=(buf[n]-'0');
  }
  return (t-32400); // Use Magic Number.
                    // TiPO has a system-time adjusted to JST(GMT+9).
}

int strncpy_until_wox(char *dest, const char *src, int n, char c){
  int m=0;
  while(m<n){
    dest[m] = src[m];
    if(dest[m] == '\0'){break;}
    if(dest[m] == c){break;}
    m++;
  }
  return m;
}

int istwgood(const char *s)
{
  char *twgood = "TWGOOD";
  for(int n=0;n<6;n++){
    if(s[n] != twgood[n]){return -1;}
  }
  return 0;
}

char *counter_inc(char *counter)
{
  counter[0] = (counter[0] + 0x01) & 0xff;
  counter[1] = (counter[1] - 0x01) & 0xff;
  return counter;
}

size_t fsize(const char *path)
{
  FILE *fp;
  size_t len=0;
  if((fp=fopen(path, "rb")) == NULL){
    fprintf(stderr, "file open error - [%s]\n", path);
    return 0;
  }
  while(fgetc(fp) != EOF){len++;}
  fclose(fp);
  return len;
}

char *strslash2hyphen(const char *s, char *t){
  while(1){
    if(*s=='/'){*t='-';}
    else{*t=*s;}
    if(*s=='\0'){break;}
    s++; t++;
  }
  return t;
}

char *strsjis_limitchrs(char *s, int l){
  int m=0, n=0;
  while(s[n]!='\0'){
    if(issjis1((unsigned char)s[n])&&issjis2((unsigned char)s[n+1])){n+=2;}
    else{n++;}
    m++;
    if(m>=l){break;}
  }
  s[n]='\0';
  return s;
}

